/* 
 * rlocated.c
 *
 * Copyright (C) 2004 Rasto Levrinc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */ 

/* rlocated is a daemon that reads rlocate dev file and copies it to the
 * rlocate db file every INTERVAL seconds.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include "pidfile.h"
#include <string.h>
#include <paths.h>
#include <fcntl.h>
#include <sys/stat.h>

#define VERSION "0.1"

#define INTERVAL 2
#define RLOCATEDB "/var/cache/locate/rlocatedb"
#define RLOCATEDEV "/dev/rlocate"
static char *PidFile = _PATH_VARRUN "rlocated.pid";
static int  no_daemon = 0; /* don't run in daemon mode */
static int  fd_dev;        /* file handle for db file  */
static int  fd_db;         /* file handle for dev file */
static char buf[BUFSIZ];   /* buffer for copy from dev file to db file */

/*
 * clean_up
 */
static void clean_up()
{
        if (fd_db != 0) {
                close(fd_db);
        }
        if (fd_dev != 0) {
                close(fd_dev);
        }
        exit(1);
}

/*
 * restart
 */
void restart(sig)
        int sig;
{
        signal(SIGCONT, restart);
        return;
}

/*
 * stop_daemon
 */
void stop_daemon(sig)
        int sig;
{
        clean_up();
        return;
}

/*
 * main
 */
int main(argc, argv) 
        int argc;
        char *argv[];
{
        auto int ch; /* option character */
        /* Parse the command-line */
        while ((ch = getopt(argc, argv, "vn")) != EOF)
                switch((char)ch)
                {
                        case 'n':       /* don't run in daemon mode */
                                no_daemon++;
                                break;
                        case 'v':
                                printf("rlocated %s\n", VERSION);
                                exit (1);
                }

        /* check the pid file and daemonize */
        if (!no_daemon)
        {
                if (!check_pid(PidFile))
                {
                        if ( fork() == 0 )
                        {
                                int num_fds = getdtablesize();
                                auto int fl;
                                for (fl = 0; fl < num_fds; ++fl)
                                {
                                        close(fl);
                                }
                                setsid();
                        }
                        else 
                                exit (0);
                }
                else
                {
                        fputs("rlocated: Already running.\n", stderr);
                        exit(1);
                }
        }
        if (!check_pid(PidFile))
        {
                if (!write_pid(PidFile))
                        clean_up();
        }
        else
        {
                fputs("rlocated: Already running.\n", stderr);
                clean_up();
        }
        /* Signals */
        for (ch= 1; ch < NSIG; ++ch)
                signal(ch, SIG_IGN);
        signal(SIGINT, stop_daemon);
        signal(SIGKILL, stop_daemon);
        signal(SIGTERM, stop_daemon);
        signal(SIGHUP, stop_daemon);
        signal(SIGCONT, restart);


        /* the main loop */
        struct flock fl = { F_WRLCK, SEEK_SET, 0, 0, 0 };
        fl.l_pid = getpid();
        auto ssize_t n;
        while (1)
        {
                /* open db file */
                if ( (fd_db = open(RLOCATEDB, O_WRONLY|O_CREAT|O_APPEND, 
                                   S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)) < 0 )
                {
                        fprintf(stderr, "%s: can't open "RLOCATEDB": %s", 
                                        argv[0], strerror(errno) );
                        return (1);
                }
                fl.l_type = F_WRLCK;
                if (fcntl(fd_db, F_SETLKW, &fl) < 0) {
                        fprintf(stderr, "%s: fcntl lock error: %s\n", 
                                        argv[0], strerror(errno) );
                }
                /* open dev file */
                if ( (fd_dev = open(RLOCATEDEV, O_RDONLY)) == -1 )
                {
                        fprintf(stderr, "%s: can't open "RLOCATEDEV
                                        ": %s\n", argv[0], strerror(errno));
                        return (1);
                }
                /* copy all blocks from dev file to db file. */
                while((n = read(fd_dev, buf, BUFSIZ)) !=0 ) 
                {
                        if (write(fd_db, buf, n) != n) {
                                fprintf(stderr, "%s: write error "RLOCATEDB
                                        ": %s\n", argv[0], strerror(errno) );
                                return (1);
                        }

                }
                if (close(fd_dev) < 0)
                        fprintf(stderr, "%s: can't close "RLOCATEDEV": %s\n", 
                                        argv[0], strerror(errno) );
                /* release lock */
                fl.l_type = F_UNLCK;
                if (fcntl(fd_db, F_SETLKW, &fl) < 0) {
                        fprintf(stderr, "%s: fcntl unlock error: %s\n", 
                                        argv[0], strerror(errno) );
                }
                if (close(fd_db) < 0) 
                        fprintf(stderr, "%s: can't close "RLOCATEDB": %s\n", 
                                        argv[0], strerror(errno) );

                sleep(INTERVAL);
        }
}


